#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import deepcopy
from roller_constant_for import BackdropStyle as bs
from roller_constant_key import Option as ok
from roller_fu import (
    clear_selection,
    color_fill_selection,
    make_layer_group,
    merge_layer_group,
    remove_z,
    select_ellipse
)
from roller_maya_style import Style
from roller_one import enumerate_name
from roller_one_the import The, get_factor_h, get_factor_w
from roller_view_hub import set_brush_details
from roller_view_real import (
    add_sub_base_group,
    add_wip_layer,
    do_gradient_for_layer,
    finish_style,
    get_point_on_edge,
    insert_copy
)
import colorsys
import gimpfu as fu

grad = None
pdb = fu.pdb


def do_diagonal(v, d, z, color):
    """
    Sample color and return a gradient's start and end points.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    z: layer
        of Backdrop Image

    color: list
        of sampled color

    Return: tuple
        ((start point, end point), ...)
    """
    a = d[ok.SAMPLE_POINTS]
    q = get_diagonal_points(v, d, a)

    for i in range(a):
        x, y = q[i]
        color[i] = pick_color(v, d, z, x, y)

    w, h = v.wip.size
    return (
        (0, q[0][0] / w),
        (0, q[0][1] / h),
        (0, q[-1][0] / w),
        (0, q[-1][1] / h)
    )


def do_horizontal(v, d, z, color):
    """
    Sample color and return a gradient's start and end points.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    z: layer
        Backdrop Image

    color: list
        of sampled color

    Return: tuple
        ((start point, end point), ...)
        Are in gradient factor.
    """
    a = d[ok.SAMPLE_POINTS]
    q = get_horizontal_points(v, d, a)

    for i in range(a):
        x, y = q[i]
        color[i] = pick_color(v, d, z, x, y)
    return (0, .0), (0, .5), (0, 1.), (0, .5)


def do_vertical(v, d, z, color):
    """
    Sample color and return a gradient's start and end points.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    z: layer
        Backdrop Image

    color: list
        of sampled color

    Return: tuple
        ((start point, end point), ...)
        Are in gradient factor.
    """
    a = d[ok.SAMPLE_POINTS]
    q = get_vertical_points(v, d, a)

    for i in range(a):
        x, y = q[i]
        color[i] = pick_color(v, d, z, x, y)
    return (0, .5), (0, .0), (0, .5), (0, 1.)


def get_diagonal_points(v, d, a):
    """
    Get the sample points for a topleft vector.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    a: int
        sample point count

    Return: list
        of sample point coordinate
        [(x, y), ...]
        Are WIP relative image coordinate.
    """
    f = d[ok.DIAGONAL_ROTATION]
    x, y = get_point_on_edge(v, f)
    x1, y1 = get_point_on_edge(v, f + 180)
    x, y = x - v.wip.x, y - v.wip.y
    x1, y1 = x1 - v.wip.x, y1 - v.wip.y
    q = []
    b = a + 1
    right_x, bottom_y = v.wip.w - 1, v.wip.h - 1

    # run, rise of slope
    w1 = (x1 - x) / b
    h1 = (y1 - y) / b

    for i in range(a):
        x = min(x + w1, right_x)
        y = min(y + h1, bottom_y)
        q.append((x, y))
    return q


def get_horizontal_points(v, d, a):
    """
    Get the sample points for a horizontal vector.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    a: int
        sample point count

    Return: list
        of sample point coordinate
        [(x, y), ...]
        Are WIP relative image coordinate.
    """
    x = 0
    f = get_factor_h(d[ok.START_Y])
    y = int(f)
    q = []
    w = v.wip.w // (a + 1)
    right_x = v.wip.w - 1

    for i in range(a):
        x = min(x + w, right_x)
        q.append((x, y))
    return q


def get_vertical_points(v, d, a):
    """
    Get the sample points for a vertical vector.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    a: int
        sample point count

    Return: list
        of sample point coordinate
        [(x, y), ...]
        Are WIP relative image coordinate.
    """
    f = get_factor_w(d[ok.START_X])
    x = int(f)
    y = 0
    q = []
    h = v.wip.h // (a + 1)
    bottom_y = v.wip.h - 1

    for i in range(a):
        y = min(y + h, bottom_y)
        q.append((x, y))
    return q


def pick_color(v, d, z, x, y):
    """
    Pick a color from the canvas.

    v: View
    d: dict
        Image Gradient Preset
        {Option key: value}

    z: layer
        the backdrop image

    x, y: int
        canvas coordinate to pick from

    Return: color
        the picked color
        RGBA
    """
    return pdb.gimp_image_pick_color(
        z.image, z,
        float(v.wip.x + x), float(v.wip.y + y),
        0,                              # no sample merged
        1,                              # yes, sample radius
        d[ok.SAMPLE_RADIUS]
    )


GET_POINTS = do_vertical, do_horizontal, do_diagonal


def make_style(v, maya):
    """
    Make a Backdrop Style layer.

    v: View
    maya: ImageGradient
    Return: layer
        with Backdrop Style material
    """
    d = maya.value_d
    parent = add_sub_base_group(v, maya)
    bg_z = insert_copy(v, parent, parent)
    a = d[ok.SAMPLE_POINTS]

    # gradient Preset dict, 'e'
    e = deepcopy(d)

    e[ok.OFFSET] = 0
    color = [0] * a
    x = bs.VECTOR.index(d[ok.SAMPLE_VECTOR])

    # Preserve.
    q = pdb.gimp_context_get_foreground()

    pdb.gimp_context_set_sample_transparent(1)

    if d[ok.PREVIEW_MODE] == bs.SHOW_SAMPLE and v.is_preview:
        show_sample(v, maya, parent, bg_z, d, a, x, color)

    else:
        show_gradient(v, parent, bg_z, d, e, a, x, color)

    # Restore.
    pdb.gimp_context_set_foreground(q)

    remove_z(bg_z)

    z = merge_layer_group(parent)

    if v.is_preview:
        z.name = "Image Gradient Sample"
        return z
    else:
        return finish_style(z, "Image Gradient")


def show_gradient(v, parent, z, d, e, a, x, color):
    """
    Draw the Image Gradient.

    v: View
    parent: layer
        destination group

    z: layer
        with backdrop image

    d: dict
        ImageGradient Preset
        {Option key: value}

    e: dict
        modified options for drawing gradient

    a: int
        sample points

    x: int
        sample vector index
    """
    global grad

    create_q = The.image_gradient_created

    if grad:
        if grad in create_q:
            create_q.pop(create_q.index(grad))

        pdb.gimp_gradient_delete(grad)
        grad = None

    gradients = The.cat.gradient_list

    # gradient name, 'n'
    n = d[ok.NAME]

    if not n:
        n = "Sampled Gradient"

    # The option name is needed in the gradient list so that
    # 'set_gimp_gradient' will pass the gradient-exists test.
    if n in gradients:
        n = enumerate_name(n, gradients)

    gradients += [n]
    grad = e[ok.GRADIENT] = pdb.gimp_gradient_new(n)

    if a > 2:
        pdb.gimp_gradient_segment_range_split_uniform(grad, 0, 0, a - 1)

    # Set 'color'.
    e[ok.START_X], e[ok.START_Y], e[ok.END_X], e[ok.END_Y] = \
        GET_POINTS[x](v, e, z, color)

    for x in range(a - 1):
        opacity = color[x].alpha * 100.

        pdb.gimp_gradient_segment_set_left_color(grad, x, color[x], opacity)

        opacity = color[x + 1].alpha * 100.
        pdb.gimp_gradient_segment_set_right_color(
            grad, x, color[x + 1], opacity
        )

    # gradient layer, 'z'
    z = do_gradient_for_layer(v, e, parent, 0)
    z.name = "Image Gradient"

    if d[ok.IRK][ok.KEEP_GRADIENT]:
        # Store image gradients so they can
        # be deleted before the program closes.
        create_q.append(grad)

        # The gradient is kept. The gradient is
        # saved to a file when GIMP closes.
        The.image_gradient_used = grad

    else:
        pdb.gimp_gradient_delete(grad)
        grad = None
    gradients.pop(-1)


def show_sample(v, maya, parent, z, d, a, x, color):
    """
    Show a preview with the sample area filled with sampled color.

    v: View
    maya: Style
    z: layer
        of Backdrop Image

    d: dict
        ImageGradient Preset

    a: int
        sample point count

    x: int
        vector index

    color: list
        of sampled color
    """
    j = z.image
    group = make_layer_group(j, "Image Gradient Sample", parent=parent)

    pdb.gimp_selection_none(j)
    pdb.gimp_context_set_foreground((0, 0, 0))

    q = (
        get_vertical_points, get_horizontal_points, get_diagonal_points
    )[x](v, d, a)

    # Set 'color'.
    GET_POINTS[x](v, d, z, color)

    w = d[ok.SAMPLE_RADIUS]
    w1 = w + w
    w2 = w + 1
    w3 = w2 * 2

    # Draw a line.
    line = q[0][0], q[0][1], q[-1][0], q[-1][1]
    line_layer = add_wip_layer(v, maya, "Line", group=group)
    sample_layer = add_wip_layer(v, maya, "Sample", group=group)
    color1 = color

    # A sample is not part of the final render, so this is okay.
    d[ok.OPACITY] = 100.
    d[ok.MODE] = "Normal"

    set_brush_details()
    pdb.gimp_selection_none(j)
    pdb.gimp_paintbrush_default(line_layer, len(line), line)

    for x1, u in enumerate(q):
        x, y = u
        r, g, b, alpha = [(i / 255.) for i in color1[x1]]
        hsv = colorsys.rgb_to_hsv(r, g, b)
        color = (0, 0, 0) if hsv[2] > .5 else (255, 255, 255)

        if alpha < 1.:
            z = add_wip_layer(v, maya, "Alpha", group=group)
            z1 = add_wip_layer(v, maya, "Alpha 2", group=group)
            z1.opacity = 100. * alpha

        else:
            z = z1 = sample_layer

        x, y = x + v.wip.x, y + v.wip.y
        select_ellipse(j, x - w2, y - w2, w3, w3)
        color_fill_selection(z, color)
        select_ellipse(j, x - w, y - w, w1, w1)

        if z1.opacity:
            color_fill_selection(z1, (r, g, b))
        if z1.opacity < 100.:
            # Remove the material as it is visible when semi-opaque.
            clear_selection(line_layer, keep_sel=True)
            clear_selection(z)
    merge_layer_group(group)


class ImageGradient(Style):
    """Create Backdrop Style output."""
    bump_row_k = ok.BRR
    is_dependent = True

    def __init__(self, *q, **d):
        """
        q: tuple
            Style spec

        d: dict
            Style spec
        """
        k_path = d['k_path']
        d['k_path'] = [k_path, k_path + (ok.IRK,)]
        Style.__init__(self, *q + (make_style,), **d)
