#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_key import BackdropStyle as by, Option as ok
from roller_def import get_default_value
from roller_fu import (
    blur_selection,
    clone_layer,
    color_fill_layer,
    make_layer_group,
    merge_layer,
    merge_layer_group
)
from roller_maya_style import Style
from roller_view_real import (
    add_sub_base_group, add_wip_layer, do_gradient_for_layer, finish_style
)
from roller_view_hub import set_fill_context, set_gimp_pattern
import gimpfu as fu

pdb = fu.pdb


def make_style(v, maya):
    """
    Make a style layer.

    v: View
    maya: SpecimenSpeckle
    Return: layer or None
        the style layer
    """
    def _do_hard_mix(_pattern):
        """
        Create a Hard Mix layer.

        _pattern: string
            pattern id

        Return: layer
            the hard mix layer
        """
        _z = clone_layer(z, n="Hard Mix")
        _z.opacity = 50.
        _z.mode = fu.LAYER_MODE_HARD_MIX

        set_gimp_pattern(_pattern)

        # x, y layer offset, '1'
        pdb.gimp_drawable_edit_bucket_fill(_z, fu.FILL_PATTERN, 1, 1)

        _do_wind(_z)
        return _z

    def _do_overlay():
        """
        Create an Overlay layer.

        Return: layer
            the overlay layer
        """
        _z = clone_layer(z, n="Overlay")
        _z.opacity = 100.
        _z.mode = fu.LAYER_MODE_OVERLAY

        pdb.gimp_drawable_invert(_z, 0)
        blur_selection(_z, 5)
        return _z

    def _do_wind(_z):
        """
        Do the wind function for each of the four directions.

        _z: layer
            to receive wind
        """
        # the direction of wind, 'i'
        for i in (0, 1, 2, 3):
            pdb.plug_in_wind(
                j, _z,
                0,              # threshold
                i,
                25,             # strength
                1,              # blast
                1               # leading edge
            )

    j = v.j
    d = maya.value_d
    group = add_sub_base_group(v, maya)
    group1 = make_layer_group(j, maya.any_group.item.key, parent=group)
    base = add_wip_layer(v, maya, "Base", group=group1)

    color_fill_layer(base, d[ok.COLOR_1])

    e = get_default_value(by.GRADIENT_FILL)
    e[ok.GRADIENT] = d[ok.GBR][ok.GRADIENT]
    e[ok.OPACITY] = d[ok.GRADIENT_OPACITY]
    z = do_gradient_for_layer(v, e, group1, 0)

    _do_overlay()

    # gradient
    e[ok.END_X] = e[ok.END_Y] = .0
    e[ok.START_X] = e[ok.START_Y] = 1.
    z = do_gradient_for_layer(v, e, group1, 0)

    set_fill_context(
        {
            ok.THRESHOLD: .6,
            ok.FILL_OPACITY: 100.,
            ok.FILL_MODE: "Normal",
            ok.CRITERION: "Composite"
        }
    )

    e = d[ok.P3R]

    # pattern #1
    z = _do_hard_mix(e[ok.PATTERN_1])
    z = _do_overlay()

    # pattern #2
    z = _do_hard_mix(e[ok.PATTERN_2])
    z = _do_overlay()

    # pattern #3
    z = _do_hard_mix(e[ok.PATTERN_3])

    _do_overlay()

    # phase two
    z = merge_layer_group(group1)
    z1 = clone_layer(z)
    z2 = clone_layer(z1)

    clone_layer(z2)
    blur_selection(z1, 6)
    blur_selection(z2, 6)

    # no linear, '0'
    pdb.gimp_drawable_invert(z1, 0)

    z = merge_layer_group(group)
    z = z1 = clone_layer(z, n="Hard Mix")
    z.mode = fu.LAYER_MODE_HARD_MIX
    z = clone_layer(z, n="LCH Color")
    z.mode = fu.LAYER_MODE_LCH_COLOR

    pdb.plug_in_colorify(j, z, d[ok.COLOR_1])
    pdb.gimp_image_merge_down(j, z1, fu.CLIP_TO_IMAGE)

    z = merge_layer(z)
    return finish_style(z, "Specimen Speckle")


class SpecimenSpeckle(Style):
    """Create Backdrop Style output."""
    bump_row_k = ok.GBR
    is_dependent = False

    def __init__(self, *q, **d):
        """
        q: tuple
            Style spec

        d: dict
            Style spec
        """
        a = d['k_path']
        d['k_path'] = [a, a + (ok.P3R,), a + (ok.GBR,), a + (ok.IRR,)]
        Style.__init__(self, *q + (make_style,), **d)
