#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import BackdropStyle as bs, Gradient as fg
from roller_constant_key import BackdropStyle as by, Option as ok
from roller_def import get_default_value
from roller_fu import (
    add_layer,
    clone_layer,
    flip_layer,
    make_layer_group,
    merge_layer_group,
    paste_layer_into
)
from roller_maya_style import Style
from roller_view_real import add_sub_base_group, clip_to_wip, finish_style
from roller_view_hub import draw_gradient, get_layer_points
import gimpfu as fu

pdb = fu.pdb


def make_style(v, maya):
    """
    Make a style layer. The main process is to draw
    a spiral gradient at the center of an image tile.
    The output is split by a potential tile function.

    v: View
    maya: SpiralChannel
    Return: layer
        the style layer
    """
    j = j1 = v.j
    d = maya.value_d
    group = add_sub_base_group(v, maya)

    # A WIP layer fails.
    z = add_layer(j, "Base", parent=group)

    e = get_default_value(by.GRADIENT_FILL)
    e[ok.GRADIENT] = "FG to BG (RGB)"

    if d[ok.GRADIENT_DIRECTION] == bs.COUNTER_CLOCKWISE:
        e[ok.GRADIENT_TYPE] = fg.SPIRAL_COUNTER_CW

    else:
        e[ok.GRADIENT_TYPE] = fg.SPIRAL_CLOCKWISE

    e[ok.BUMP] = get_default_value(ok.BUMP)
    e[ok.START_X] = e[ok.START_Y] = e[ok.END_Y] = .5
    e[ok.END_X] = .5 + d[ok.SPIRAL_DISTANCE]
    color = d[ok.COLOR_1]

    # The color inverts when using the difference mode.
    # This reverses that.
    if d[ok.SPIRAL_MOD] != "None":
        color = tuple([255 - i for i in d[ok.COLOR_1]])

    # Preserve.
    q = pdb.gimp_context_get_foreground()
    q1 = pdb.gimp_context_get_background()

    pdb.gimp_context_set_foreground(color)
    pdb.gimp_context_set_background((255, 255, 255))

    is_tile = False
    left_x, top_y, w, h = v.wip.rect
    row, column = d[ok.ROW], d[ok.COLUMN]

    if row > 1 or column > 1:
        is_tile = True
        w = w // column
        h = h // row
        j1 = pdb.gimp_image_new(w, h, fu.RGB)
        z = add_layer(j1, "Tile")
        pdb.gimp_image_set_active_layer(j1, z)

    x, y, x1, y1 = get_layer_points(e, w, h)

    if not is_tile:
        x += left_x
        y += top_y
        x1 += left_x
        y1 += top_y

    z = draw_gradient(z, e, x, y, x1, y1)

    if d[ok.SPIRAL_MOD] != "None":
        z = clone_layer(z)
        z.mode = fu.LAYER_MODE_DIFFERENCE

        if d[ok.SPIRAL_MOD] in (bs.HORIZONTAL_FLIP, bs.FLIP_BOTH):
            flip_layer(z, horizontal=1)

        if d[ok.SPIRAL_MOD] in (bs.VERTICAL_FLIP, bs.FLIP_BOTH):
            flip_layer(z)

    if is_tile:
        pdb.gimp_edit_copy_visible(j1)
        pdb.gimp_image_delete(j1)

        tiles = make_layer_group(j, "Tiles", parent=group)
        x = left_x
        y = top_y

        # column tiling
        z = first = paste_layer_into(tiles)
        pdb.gimp_image_reorder_item(j, z, tiles, 0)

        for c in range(column):
            if c:
                z = clone_layer(first)
                x = (c * w) + left_x

            if c % 2:
                flip_layer(z, horizontal=1)
            pdb.gimp_layer_set_offsets(z, x, y)

        z = first = merge_layer_group(tiles)

        # row tiling
        if row > 1:
            tiles = make_layer_group(j, "Tiles", parent=group, z=z)
            x = left_x

            for r in range(row - 1):
                z = clone_layer(first)

                if not r % 2:
                    flip_layer(z)

                y += h
                pdb.gimp_layer_set_offsets(z, x, y)

            z = merge_layer_group(tiles)
        if z.width != v.wip.w or z.height != v.wip.h:
            w = ((v.wip.w - z.width) // 2) + left_x
            h = ((v.wip.h - z.height) // 2) + top_y
            pdb.gimp_layer_set_offsets(z, w, h)

    z = merge_layer_group(group)

    # Restore.
    pdb.gimp_context_set_foreground(q)
    pdb.gimp_context_set_background(q1)

    clip_to_wip(v, z)
    return finish_style(z, "Spiral Channel")


class SpiralChannel(Style):
    """Create Backdrop Style output."""
    bump_row_k = ok.BRR
    is_dependent = False

    def __init__(self, *q, **d):
        """
        q: tuple
            Style spec

        d: dict
            Style spec
        """
        Style.__init__(self, *q + (make_style,), **d)
