#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_key import Option as ok
from roller_frame import Noise, draw_color_profile
from roller_frame_build import Build
from roller_fu import add_layer, clone_layer, select_item
from roller_fu_mode import get_mode
from roller_maya import check_matter, make_frame_group, check_frame_cake
from roller_maya_blur_behind import BlurBehind
from roller_maya_light import Light
from roller_maya_shadow import Shadow
from roller_view_hub import PROFILE
from roller_view_real import LIGHT, get_light, get_noise
import gimpfu as fu

pdb = fu.pdb


def adjust_opacity(z):
    """
    For a given layer's RGB pixel, adjust its alpha channel
    mean intensity value to be above a threshold (225).

    z: layer
        with material to adjust

    Return: layer, Selection state
        adjusted
    """
    j = z.image

    # the mean intensity value, 'f'
    f = 0

    while f < 225.:
        # Create a selection-mask.
        select_item(z)

        if not pdb.gimp_selection_is_empty(j):
            # Get the mean intensity value of the material's alpha, 'f'.
            f = pdb.gimp_drawable_histogram(z, fu.HISTOGRAM_ALPHA, .0, 1.)[0]
            if f < 225.:
                # double-down to double the alpha value
                z = clone_layer(z)
                z = pdb.gimp_image_merge_down(j, z, fu.CLIP_TO_IMAGE)
        else:
            break
    return z


def do_matter(v, maya):
    """
    Make a frame.

    v: View
    maya: Maya
    Return: layer or None
        with a frame
    """
    j = v.j
    d = maya.value_d
    group = maya.group
    z = add_layer(
        j,
        group.name + " Color Pipe",
        parent=group,
        offset=get_light(maya) + get_noise(maya)
    )
    color = d[ok.COLOR_1]
    w = d[ok.FRAME_W]
    q = PROFILE[d[ok.PROFILE]](w, *(color,))

    select_item(maya.cause.matter)
    draw_color_profile(z, w, q, color)

    z = adjust_opacity(z)
    z.opacity = d[ok.FRAME_OPACITY]
    z.mode = get_mode(d, k=ok.FRAME_MODE)
    return z


class ColorPipe(Build):
    """Create a frame with a custom profile and color."""
    issue_q = 'cake', 'matter', 'shade'
    put = (
        (make_frame_group, 'group'),
        (check_matter, 'matter'),
        (check_frame_cake, None)
    )

    def __init__(self, any_group, super_maya, k_path=()):
        """
        any_group: AnyGroup
            Owns the option.

        super_maya: Maya
            super type
            Is from the AnyGroup where the Frame Widget resides.

        k_path: tuple
            Key path of the Frame Button in its vote dict.
            (Option key, ...)
        """
        self.do_matter = do_matter

        Build.__init__(self, any_group, super_maya, k_path)

        self.sub_maya[ok.NOISE_D] = Noise(
            any_group, self, k_path + (ok.NSR, ok.NOISE_D)
        )
        self.sub_maya[ok.SHADOW] = Shadow(
            any_group, self, (self.cause, self), k_path + (ok.NSR, ok.SHADOW)
        )
        self.sub_maya[ok.BLUR_BEHIND] = BlurBehind(any_group, self, k_path)
        self.sub_maya[LIGHT] = Light(any_group, self, ok.TRANSLUCENT)

    def do(self, v, d, is_change):
        """
        Check, modify, and produce layer output.

        v: View
        d: dict
            Color Pipe Preset
            {Option key: value}

        is_change: bool
            Is the state of the super Maya's matter and/or mask.

        Return: bool
            If True, then Shadow changed the background.
        """
        self.value_d = d
        is_back = v.is_back
        self.is_matter |= is_change

        self.realize(v)
        self.sub_maya[ok.NOISE_D].do(v, d[ok.NSR][ok.NOISE_D], self.is_matter)

        m = self.sub_maya[ok.SHADOW].do(
            v,
            d[ok.NSR][ok.SHADOW],
            self.is_matter + self.is_cake,
            is_change + self.cause.is_cake
        )

        self.sub_maya[ok.BLUR_BEHIND].do(v, d, m or is_back, self.is_matter)
        self.sub_maya[LIGHT].do(v, self.is_matter)
        self.reset_issue()
        return m
