#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import (
    Frame as ff, Plan as fy, Model as mo, Signal as si
)
from roller_constant_key import (
    Frame as ek, Node as ny, Option as ok, Plan as ak
)
from roller_deco import make_deco_mask
from roller_deco_image import (
    do_main_cell,
    do_main_face,
    do_canvas,
    do_cell,
    do_face,
    i_make_main_face_mask,
    i_make_cell_face_mask,
    mask_main_cell
)
from roller_maya import (
    MAIN,
    PER,
    CanvasRoute,
    CellRoute,
    ImageRoll,
    assign_mask_image,
    check_cake,
    check_matter
)
from roller_maya_blur_behind import BlurBehind
from roller_maya_light import Light
from roller_maya_mask import Mask
from roller_maya_plan_detail import Detail
from roller_maya_plan_name import (
    draw_main_name,
    draw_canvas_name,
    draw_cell_name,
    draw_face_main_name,
    draw_face_cell_name
)
from roller_one_extract import get_option_list_choice
from roller_one_image import Image as ig
from roller_option_group import DecoGroup
from roller_view_option_list import Frame
from roller_view_real import (
    LIGHT, make_canvas_group, make_cast_group, remove_maya_z
)
from roller_view_step import get_planner
import gimpfu as fu

pdb = fu.pdb


def assign_canvas_image(v, maya, d):
    """
    Assign image for the Canvas-branch.
    v: View
    maya: Maya
    d: dict
        Image Preset
    """
    j, j1, j2 = assign_preset_source(v, d)

    if (
        j != maya.get_image(None) or
        j1 != maya.get_mask_image(None) or
        j2 != maya.get_frame_image(None)
    ):
        maya.is_matter = True

    maya.set_image(None, j)
    maya.set_mask_image(None, j1)
    maya.set_frame_image(None, j2)


def assign_image(v, maya, p):
    """
    Assign image for the Cell-branch. Has three image for each cell.

    v: View
    maya: Maya
    p: function
        Assign an Image, a Mask Image, and a Frame Image.
    """
    d = maya.value_d

    per_cell = d[ok.PER_CELL]
    cell_d = maya.cell_d
    for r_c in maya.model.cell_q:
        if r_c in per_cell:
            e = per_cell[r_c]
            cell_maya = cell_d[r_c]

        else:
            e = d
            cell_maya = maya

        image, mask, frame = p(v, e)

        # Determine if an image reference changed.
        if image != cell_maya.get_image(r_c):
            cell_maya.is_matter = True

        if mask != cell_maya.get_mask_image(r_c):
            cell_maya.sub_maya[ok.MASK].is_matter = True

        if frame != cell_maya.get_frame_image(r_c):
            cell_maya.sub_maya[ok.FRAME].is_matter = True

        # Store image reference.
        cell_maya.set_image(r_c, image)
        cell_maya.set_mask_image(r_c, mask)
        cell_maya.set_frame_image(r_c, frame)


def assign_cell_image(v, maya):
    """
    Assign image for a cell.

    v: View
    maya: Maya
    """
    def _assign(_v, _d):
        """
        _v: View
        _d: dict
            Image Preset
            {Option key: value}

        Return: tuple
            (Image, Mask Image, Frame Image)
            any could be None
        """
        return assign_preset_source(_v, _d)
    assign_image(v, maya, _assign)


def assign_face_image(v, maya):
    """
    Assign Cell/Face image.

    v: View
    maya: Maya
    """
    def _assign(_v, _d):
        """
        _v: View
        _d: dict
            Image Preset
            {Option key: value}

        Return: tuple
            (Image, Mask Image, Frame Image)
            any could be None
        """
        _image_q = []
        _mask_q = []

        for _i in range(3):
            _j, _j1 = assign_preset_source(_v, _d, is_face=True)[:2]

            _image_q.append(_j)
            _mask_q.append(_j1)
        return _image_q, _mask_q, None
    assign_image(v, maya, _assign)


def assign_preset_source(v, d, is_face=False):
    """
    Assign an image, mask image, and frame image for a Preset.

    v: View
    d: dict
        Image Preset

    is_face: bool
        Is True when a Face Maya is calling.

    Return: tuple
        (Image or None, Image or None, Image or None)
        (the image, the mask image, the frame image)
    """
    def _nail_polish():
        if sub_frame_d[ok.FRAME_TYPE] == ff.IMAGE:
            return ig.get_image(
                sub_frame_d[ok.IPB][ok.IMAGE_CHOICE],
                v.image_source
            )

    def _frame_over():
        if sub_frame_d[ok.FRAME_STYLE] == ff.IMAGE:
            return ig.get_image(
                sub_frame_d[ok.CIR][ok.IMAGE_CHOICE],
                v.image_source
            )

    e = d[ok.IRRW][ok.IMAGE_CHOICE]
    go = d[ok.SWITCH] and e[ok.SWITCH]
    a = ig.get_image(e, v.image_source) if go else None

    # If the primary image is None, then the
    # mask and frame are None, 'b, c'.
    b = c = None

    if a:
        b = assign_mask_image(v, d[ok.MRR][ok.MASK]) if go else None

        # Face doesn't have a Frame option.
        if not is_face:
            # Determine if Frame Over has an image assignment.
            frame_d = d[ok.MRR][ok.FRAME]
            if frame_d[ok.SWITCH]:
                option_k, sub_frame_d = get_option_list_choice(frame_d)
                p = {
                    ek.FRAME_OVER: _frame_over, ek.NAIL_POLISH: _nail_polish
                }.get(option_k)
                if p:
                    c = p()
    return a, b, c


class Image(DecoGroup):
    """Manage Image layer output."""

    def __init__(self, **d):
        DecoGroup.__init__(self, **d)

        node_k = self.step_key[-2]
        self.plan = {
            ny.CANVAS: PlanCanvas, ny.CELL: PlanCell, ny.FACE: PlanFace
        }[node_k](self)
        self.work = {
            ny.CANVAS: WorkCanvas, ny.CELL: WorkCell, ny.FACE: WorkFace
        }[node_k](self)
        baby = self.item.model.baby
        if node_k == ny.FACE:
            self.handle_d[
                baby.connect(si.CELL_SHIFT_CALC, self.on_cell_calc)
            ] = baby


class Chi(ImageRoll):
    """Is factored from Plan and Work."""
    MASK_ROW_K = ok.MRR

    def __init__(
        self, any_group, view_x, make_mask, get_mask_d, q
    ):
        """
        view_x: int
            0 or 1
            Plan or Work index

        make_mask: function
            Call to make Mask for the matter layer.

        q: iterable
            Direct output.
        """
        ImageRoll.__init__(
            self,
            any_group,
            view_x,
            q,
            k_path=[
                (),
                (ok.IRRW, ok.IMAGE_CHOICE),
                (ok.IRRW, ok.RESIZE),
                (ok.FLIP_R,)
            ]
        )
        self.set_issue()
        self.sub_maya[ok.MASK] = Mask(
            any_group, self, view_x, make_mask, get_mask_d, (ok.MRR, ok.MASK)
        )


class Plan(Chi):
    """Manage layer output for Draft and Plan views."""
    put = (make_cast_group, 'group'), (check_matter, 'matter')

    def __init__(self, any_group, make_mask, get_mask_d, draw_name):
        """
        any_group: AnyGroup
        make_mask: function
            Call to make a mask for the matter layer.

        get_mask_d: function
            Call to retrieve the Mask Preset.

        draw_name: function or None
            Call to produce an image name text layer.
        """
        Chi.__init__(self, any_group, 0, make_mask, get_mask_d, self.put)

        planner = get_planner(any_group.step_key)
        self.is_planned = planner.get_option_a(ak.IMAGE)

        self.handle_d[
            planner.connect(fy.SIGNAL_D[ak.IMAGE], self.on_plan_option_change)
        ] = planner
        if draw_name:
            self.sub_maya[ak.NAME] = Detail(
                any_group, self, draw_name, ak.NAME
            )

    def bore(self, v):
        """
        Manage layer output during a View run.

        v: View
        """
        d = self.value_d
        self.go = d[ok.SWITCH]

        if self.go:
            if self.is_planned:
                self.is_matter |= self.is_switched

            else:
                # Produce a layer group and nothing else.
                # Plan Detail/name requires the layer group.
                remove_maya_z(self, 'matter')
                self.is_matter = False

        self.realize(v)
        if self.go:
            self.go = bool(self.matter)

            self.sub_maya[ok.MASK].do(v, d[ok.MRR][ok.MASK])
            if ak.NAME in self.sub_maya:
                self.sub_maya[ak.NAME].do(v)

    def on_plan_option_change(self, _, arg):
        """Respond to change in the Planner's Image checkbox."""
        self.is_planned, self.is_switched = arg


class Work(Chi):
    """Manage Peek, Preview, and render view layer output."""
    put = (
        (make_cast_group, 'group'),
        (check_matter, 'matter'),
        (check_cake, None)
    )

    def __init__(self, any_group, make_mask, get_mask_d):
        """
        v: View
        make_mask: function
            Call to make a mask layer.

        get_mask_d: function
            Is a callback to fetch a Mask Preset value dict.
        """
        Chi.__init__(self, any_group, 1, make_mask, get_mask_d, self.put)

        self.sub_maya[ok.FRAME] = Frame(any_group, self, (ok.MRR, ok.FRAME))
        self.sub_maya[ok.BLUR_BEHIND] = BlurBehind(any_group, self, ())
        self.sub_maya[LIGHT] = Light(any_group, self, ok.IMAGE)

    def bore(self, v):
        """
        Manage layer output during a View run.

        v: View
        """
        d = self.value_d
        self.go = d[ok.SWITCH]
        is_back = v.is_back

        self.realize(v)
        if self.go:
            self.go = bool(self.matter)
            is_change = self.is_matter or self.sub_maya[ok.MASK].is_matter

            if is_change:
                self.sub_maya[ok.MASK].do(v, d[ok.MRR][ok.MASK])

            if not self.is_face:
                is_back |= self.sub_maya[ok.FRAME].do(
                    v, d[ok.MRR][ok.FRAME], is_change
                )

            self.sub_maya[ok.BLUR_BEHIND].do(v, d, is_back, is_change)
            self.sub_maya[LIGHT].do(v, is_change)


class Main:
    """Is factored from Plan and Work for the Main window option settings."""
    vote_type = MAIN

    def __init__(self):
        self.r_c = None

    def get_mask_d(self):
        return self.any_group.value_d[ok.MRR][ok.MASK]


class WorkMain(Main, Work):
    """Manage layer output for the View/Work main option settings."""

    def __init__(self, *arg, **kwarg):
        """
        arg: tuple
            Work spec

        kwarg: dict
            Work spec
        """
        Main.__init__(self)
        Work.__init__(self, *arg + (self.get_mask_d,), **kwarg)


# Canvas_______________________________________________________________________
class Cloth:
    """Is factored from Plan and Work Canvas Maya."""

    def __init__(self):
        self.do_matter = do_canvas

    def prep(self, v):
        """
        For every View, there is an Image. Assign image for a Canvas-branch.

        v: View
        """
        assign_canvas_image(v, self, self.value_d)


class PlanCanvas(Main, Plan, CanvasRoute, Cloth):
    """Manage Plan/Canvas simulated layer output."""
    issue_q = 'matter', 'switched'
    put = (make_canvas_group, 'group'), (check_matter, 'matter')

    def __init__(self, any_group):
        Main.__init__(self)
        Plan.__init__(
            self,
            any_group,
            make_deco_mask,
            self.get_mask_d,
            draw_canvas_name
        )
        CanvasRoute.__init__(self)
        Cloth.__init__(self)


class WorkCanvas(WorkMain, CanvasRoute, Cloth):
    """Manage Work/Canvas layer output."""
    issue_q = 'cake', 'matter', 'shade'
    put = (
        (make_canvas_group, 'group'),
        (check_matter, 'matter'),
        (check_cake, None)
    )

    def __init__(self, any_group):
        self.do_matter = do_canvas

        # for error test
        WorkMain.__init__(self, any_group, make_deco_mask)
        CanvasRoute.__init__(self)
        Cloth.__init__(self)
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯


# Cell_________________________________________________________________________
class Cellular:
    """Is factored from Plan and Work Cell Maya."""
    is_face = False

    def __init__(self):
        self.r_c = None
        self.do_matter = do_main_cell

    def prep(self, v):
        """
        For every View, there is an Image.

        v: View
        """
        assign_cell_image(v, self)


class PlanCell(Main, Plan, CellRoute, Cellular):
    """Manage Plan Cell output."""
    issue_q = 'matter', 'per_cell', 'switched'

    def __init__(self, any_group):
        Main.__init__(self)
        Plan.__init__(
            self, any_group, mask_main_cell, self.get_mask_d, draw_main_name
        )
        CellRoute.__init__(self, PlanCellPer)
        Cellular.__init__(self)


class WorkCell(WorkMain, CellRoute, Cellular):
    """Manage Work Cell output."""
    issue_q = 'cake', 'matter', 'per_cell', 'shade'

    def __init__(self, any_group):
        WorkMain.__init__(self, any_group, mask_main_cell)
        CellRoute.__init__(self, WorkCellPer)
        Cellular.__init__(self)
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯


# Per Cell_____________________________________________________________________
class PerCell:
    """Is factored from Plan and Work Per Cell Maya."""
    vote_type = PER
    cell_type = mo.PER_CELL

    def __init__(self, do_matter, r_c):
        self.do_matter = do_matter
        self.r_c = r_c

    def get_mask_d(self):
        d = self.per_cell_group.get_cell_value(self.r_c)
        if d:
            return d[ok.MRR][ok.MASK]


class PlanCellPer(PerCell, Plan):
    """Manage Plan Per Cell output."""
    is_face = False
    issue_q = 'matter', 'switched'

    def __init__(self, any_group, r_c):
        PerCell.__init__(self, do_cell, r_c)
        Plan.__init__(
            self, any_group, make_deco_mask, self.get_mask_d, draw_cell_name
        )


class WorkCellPer(PerCell, Work):
    """Manage Work Per Cell output."""
    is_face = False
    issue_q = 'cake', 'matter', 'shade'

    def __init__(self, any_group, r_c):
        """
        any_group: AnyGroup
            Has the Per Cell option.

        r_c: tuple
            (row, column)
            cell index and Goo key
        """
        PerCell.__init__(self, do_cell, r_c)
        Work.__init__(self, any_group, make_deco_mask, self.get_mask_d)
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯


# Face_________________________________________________________________________
class Face:
    """Is factored from Plan and Work Face Maya."""
    is_face = True

    def __init__(self):
        self.r_c = self.r_c_x = None
        self.do_matter = do_main_face

    def prep(self, v):
        """
        For every View, there is an Image.

        v: View
        """
        assign_face_image(v, self)


class PlanFace(Face, Main, Plan, CellRoute):
    """Manage Plan Face output."""
    issue_q = 'matter', 'per_cell', 'switched'

    def __init__(self, any_group):
        self.do_matter = do_main_face

        Face.__init__(self)
        Main.__init__(self)
        Plan.__init__(
            self,
            any_group,
            i_make_main_face_mask,
            self.get_mask_d,
            draw_face_main_name
        )
        CellRoute.__init__(self, PlanFacePer)


class WorkFace(Face, WorkMain, CellRoute):
    """Manage Work Face output."""
    issue_q = 'cake', 'matter', 'per_cell', 'shade'

    def __init__(self, any_group):
        self.do_matter = do_main_face

        Face.__init__(self)
        WorkMain.__init__(self, any_group, i_make_main_face_mask)
        CellRoute.__init__(self, WorkFacePer)
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯


# Face Per Cell________________________________________________________________
class FacePer(PerCell):
    """Is factored from Plan and Work Per Cell Face Maya."""
    is_face = True

    def __init__(self, *q):
        PerCell.__init__(self, *q)


class PlanFacePer(FacePer, Plan):
    """Manage Per cell Face output."""
    issue_q = 'matter', 'switched'

    def __init__(self, any_group, r_c):
        """
        r_c: tuple
            (r, c); cell index; of int
        """
        FacePer.__init__(self, do_face, r_c)
        Plan.__init__(
            self,
            any_group,
            i_make_cell_face_mask,
            self.get_mask_d,
            draw_face_cell_name
        )


class WorkFacePer(FacePer, Work):
    """Manage Per cell Face output."""
    issue_q = 'cake', 'matter', 'shade'

    def __init__(self, any_group, r_c):
        """
        Create a Maya for Per Cell.

        r_c: tuple
            (r, c); cell index; of int
        """
        FacePer.__init__(self, do_face, r_c)
        Work.__init__(self, any_group, i_make_cell_face_mask, self.get_mask_d)
# ¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
