#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import Grid as gr, Shape as sh
from roller_model_goo import Goo
from roller_one_rect import Rect
from roller_polygon import calc_pin_xy

# The odd row and column intersects are offset in their intersect list.
ODD_INTERSECT_OFFSET = 2
OCTAGON_REMAIN = 1. - sh.OCTAGON_RATIO


class GridOctagonDouble:
    """
    Calculate the points for the shape of cells. The cells are
    octagon shaped and are arranged in a double-spaced cell model.
    The sides of the octagon appear to be facing side-to-side.
    """

    @staticmethod
    def calc(model, o):
        """
        Calculate cell rectangle and form polygon for a Model's cell.

        model: Model
        o: One
            with Cell Type reference

        Return: list
            [Plan vote, Work vote]
        """
        def _calc_intersects():
            """
            Calculate vertical and horizontal intersect points.

            Return: tuple
                grid size
            """
            _w1 = w * sh.OCTAGON_RATIO
            _h1 = h * sh.OCTAGON_RATIO
            _w2 = w - _w1
            _h2 = h - _h1
            _w3 = w - _w1 - _w1
            _h3 = h - _h1 - _h1
            _s1 = (column - 1) * _w2 + w, (row - 1) * _h2 + h
            _x, _y = calc_pin_xy(
                o.pin_corner, x, y, canvas_w, canvas_h, *_s1
            )
            _s1 = canvas_w + _x, canvas_h + _y

            # remainder total, '_f_x, _f_y'
            _f_x = _f_y = .0

            for _ in range(row):
                _y, _f = divmod(_y, 1.)
                _f_y += _f

                if _f_y >= .999:
                    _y += 1.
                    _f_y -= 1.

                q_y.extend((_y, round(_y + _h1), round(_y + _h2), round(_y + h)))
                _y += h + _h3

            for _ in range(column):
                _x, _f = divmod(_x, 1.)
                _f_x += _f

                if _f_x >= .999:
                    _x += 1.
                    _f_x -= 1.

                q_x.extend((_x, round(_x + _w1), round(_x + _w2), round(_x + w)))
                _x += w + _w3
            return _s1

        def _get_intersects():
            """
            Get the points for a double-spaced octagon.

            Return: tuple
                of octagon offsets
            """
            _c1 = c // 2 * 4
            _r1 = r // 2 * 4

            if c % 2:
                _index_x = (c - 1) * 4 + 2 - _c1

            else:
                _index_x = c * 4 - _c1

            if r % 2:
                _index_y = (r - 1) * 4 + 2 - _r1
            else:
                _index_y = r * 4 - _r1
            return q_x[_index_x:_index_x+4], q_y[_index_y:_index_y+4]

        vote_d = {}
        did_cell = model.past.did_cell
        row, column = model.division
        goo_d = model.goo_d
        x, y, canvas_w, canvas_h = model.canvas_pocket.rect

        # intersect points
        q_x, q_y = [], []

        if o.grid_type == gr.CELL_SIZE:
            # Correct cell size overflow.
            w, h = min(canvas_w, o.column_width), min(canvas_h, o.row_height)

        elif o.grid_type == gr.SHAPE_COUNT:
            w = canvas_w / (sh.OCTAGON_RATIO + column * OCTAGON_REMAIN)
            h = canvas_h / (sh.OCTAGON_RATIO + row * OCTAGON_REMAIN)
            w = min(w, h)
            w, h = w, w

        else:
            w = canvas_w / (sh.OCTAGON_RATIO + column * OCTAGON_REMAIN)
            h = canvas_h / (sh.OCTAGON_RATIO + row * OCTAGON_REMAIN)

        _calc_intersects()

        for r_c in model.cell_q:
            r, c = r_c
            q, q1 = _get_intersects()
            a = goo_d[r_c] = Goo(r_c)
            x, y = q[0], q1[0]
            x1, y1 = q[3], q1[3]

            # Prevent round to zero with max 1.
            a.cell = a.merged = Rect(x, y, max(1., x1 - x), max(1., y1 - y))

            a.form = (
                q[1], y,
                q[2], y,
                x1, q1[1],
                x1, q1[2],
                q[2], y1,
                q[1], y1,
                x, q1[2],
                x, q1[1]
            )
            vote_d[r_c] = did_cell(r_c)
        return vote_d
