#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant_for import Preset as fp, Widget as fw
from roller_constant_key import Widget as wk
from roller_port_process import PortProcess
from roller_fu_comm import pop_up, show_err
from roller_one_oz import get_preset_path
from roller_one_the import The
from roller_widget_box import Box
from roller_widget_button import AcceptButton, Button, CancelButton
from roller_widget_tree import ChoiceList
import glob
import gtk
import os


def collect_preset(group_key):
    """
    Create a list of internal and external Presets.

    group_key: string
        option group key

    Return: tuple
        (external Preset dict, Preset name list)
    """
    # list of Presets, 'q'
    q = [fp.DEFAULT]

    ext_files = []
    go = False
    external_preset = {}

    try:
        search_path = get_preset_path(
            group_key,
            u"*",
            The.preset_folder
        )

        # Ignore sub-directories.
        ext_files = glob.glob(search_path)
        go = True

    except Exception as ex:
        show_err(ex)
        show_err("Roller was unable to load presets.")

    if go:
        # Make an external Preset file list.
        for n in ext_files:
            name = os.path.splitext(
                os.path.basename(n).split(fp.PRESET_SEPARATOR)[1]
            )[0]
            external_preset[name] = n
            q.append(name)
    return external_preset, q


class PortPreset(PortProcess):
    """Is a Port for loading, saving, and deleting a Preset file."""
    window_key = "Preset"

    def __init__(self, d, g):
        """
        d: dict
            Has init values.

        g: Button
            Is responsible.

        k: string
            Preset key
        """
        self.any_group = g.any_group
        self.item_key = g.any_group.item.key
        self._external_preset = {}
        self._dependent_button = []
        self._del_button = self._load_button = self._choice_list = None
        PortProcess.__init__(self, d, g)

    def _draw_list_group(self, g):
        """
        Draw a ChoiceList containing a Preset list.

        g: GTK container
            for list group
        """
        self._external_preset, self._preset_list = collect_preset(
            self.item_key
        )
        self._choice_list = ChoiceList(
            **{
                wk.CHOICE: "Default",
                wk.FUNCTION: self.get_preset_list,
                wk.RELAY: [self.on_port_change],
                wk.ROLLER_WIN: self.roller_win,
                wk.TREE_COLOR: self.color
            }
        )

        g.pack_start(self._choice_list, expand=True)
        self.on_port_change(self._choice_list)

    def _get_selection(self):
        """
        Get the selection in the ChoiceList.

        Return: int or None
            0 to n
        """
        if self._choice_list:
            return self._choice_list.get_sel_x()

    def delete_preset(self, _):
        """
        Delete the selected Preset.

        _: Button
            the delete Button
            not used
        """
        x = self._get_selection()
        if x is not None:
            n = self._choice_list.get_a()
            if n in self._external_preset:
                path = self._external_preset[n]
                if pop_up(
                    self.roller_win.gtk_win,        # Use dialog.
                    0,
                    "Do you want to delete:\n{}?".format(path),
                    "Delete a File Confirmation"
                ):
                    try:
                        os.remove(path)
                        pop_up(
                            self.roller_win.gtk_win,        # Use dialog.
                            1,
                            "The file:\n" + path + "\nwas removed.",
                            "File Deleted"
                        )
                        self._external_preset.pop(n)
                        self._choice_list.remove_x(x)
                        self.on_port_change(self._choice_list)
                    except Exception as ex:
                        show_err(ex)
                        show_err("Roller was unable to delete the file.")

    def draw_process_group(self, g):
        """
        Draw a Widget group having process Buttons.

        g: GTK container
            to receive group
        """
        w = fw.MARGIN
        w1 = w // 2
        hbox = Box(box=gtk.HBox, align=(0, 0, 1, 1))
        cancel_button = CancelButton(
            align=(0, 0, 1, 0),
            padding=(w, w, w, w1),
            relay=[],
            roller_win=self.roller_win,
        )
        self._del_button = Button(
            align=(0, 0, 1, 0),
            padding=(w, w, w1, w1),
            relay=[self.delete_preset],
            roller_win=self.roller_win,
            text="Delete"
        )
        self._load_button = AcceptButton(
            align=(0, 0, 1, 0),
            padding=(w, w, w1, w),
            relay=[],
            roller_win=self.roller_win
        )
        q = cancel_button, self._del_button, self._load_button
        self._dependent_button = self._del_button, self._load_button

        for i in q:
            hbox.add(i)

        g.pack_end(hbox, expand=False)
        self.keep(q)

    def draw(self):
        """Draw Widget."""
        self.draw_list(
            (self._draw_list_group, self.draw_process_group),
            ("Available {}".format(self.item_key), "")
        )
        self.roller_win.gtk_win.vbox.set_size_request(350, 350)

    def get_preset_list(self):
        return self._preset_list

    def load_preset(self, *_):
        """Use to get the value of the list selection."""
        x = self._choice_list.get_sel_x()
        if x is not None:
            any_group = self.any_group
            preset = any_group.get_preset_g()

            self.roller_win.gtk_win.hide()
            preset.load_file(self._choice_list.get_a(), self.item_key)

    def get_group_value(self):
        """
        Load the selected Preset.

        Return: True
            for GTK
        """
        return self.load_preset()

    def get_hook(self):
        """
        Fetch the Port's cancel and accept responders.

        Return: tuple
            (function, function) -> (cancel hook, accept hook)
        """
        return lambda: None, self.load_preset

    def on_port_change(self, g, *_):
        """
        Respond to a list choice.

        g: ChoiceList
            with a Preset-name list
        """
        q = self._dependent_button
        x = self._get_selection()

        if x is not None:
            for i in q:
                i.enable()
        else:
            for i in q:
                i.disable()
