#!/usr/bin/env python
# -*- coding: utf-8 -*-
from copy import deepcopy
from roller_constant_for import Signal as si, Issue as vo
from roller_constant_key import (
    Button as bk,
    Group as gk,
    Option as ok,
    Step as sk,
    Widget as wk
)
from roller_one_the import The
from roller_one_extract import get_option_list_choice
from roller_one_tip import Tip
from roller_port_per_cell import PortPerCell
from roller_widget import set_widget_attr
from roller_widget_button import Button
from roller_widget_check_button import CheckButton
from roller_widget_voter import accept_vote
import gobject
import gtk


def compare_both_value(g):
    """
    Compare both the Plan and View values.

    g: PerCellGroup
        with value table
    """
    # Do Work and Plan, '2'.
    for x in range(2):
        compare_value_table(g, x)


def compare_preset_table(g, new_d, x):
    """
    Compare the Preset value table with a Preview table.

    g: PerCellGroup
        with value table

    new_d: dict
        Is the incoming value for the Per Cell Group.

    x: int
        index to Plan or Work; 0 or 1
    """
    old_d = g.view_value[x]
    is_change = new_d != old_d

    # (row, column) cell index, 'r_c'; Preset value dict, 'd'
    for r_c, d in new_d.items():
        vote_d = {}
        reap_vote(
            vote_d,
            g.default_ballot_d,
            old_d[r_c] if r_c in old_d else None,
            d
        )
        g.emit(si.PER_CELL_CHANGE, (r_c, vote_d, x))
    return is_change


def compare_type_table(g, new_d, x):
    """
    Compare the Table Cell Type value table with a Preview table.

    g: PerCellGroup
        with value table

    new_d: dict
        incoming Table's Cell-Type Per-Cell value

    x: int
        index to Plan or Work
    """
    old_d = g.view_value[x]
    is_change = old_d != new_d

    for r_c, a in new_d.items():
        if a != old_d[r_c] if r_c in old_d else None:
            m = True

        else:
            m = False
        g.emit(si.PER_CELL_CHANGE, (r_c, m, x))
    return is_change


def compare_value_table(g, x):
    """
    Compare a value table with a Preview table.
    Route Table Cell Type differently.

    g: PerCellGroup
        with value table

    x: int
        index to Plan or Work attribute
    """
    d = g.get_a()

    if g.is_type_table:
        is_change = compare_type_table(g, d, x)

    else:
        is_change = compare_preset_table(g, d, x)

    # Per Cell Widget changed
    g.any_group.cast_vote(x, ok.PER_CELL, vo.PER_CELL, is_change)

    g.any_group.cast_vote(
        x, ok.IS_MAIN, vo.BAKE, d.keys() != g.view_value[x].keys()
    )


def reap_vote(vote_d, ballot_d, old_d, new_d):
    """
    Compare two Preset values. Have each option cast a vote in a
    vote dict. The option's vote is True if the option has changed.

    vote_d: dict
        Use to collect vote string-type.

    ballot_d: dict
        {option key: votes}

    old_d: dict
        old value

    new_d: dict
        new value

    Return: dict
        with vote
    """
    if old_d:
        for i, a in ballot_d.items():
            if i == ok.FRAME:
                if i not in vote_d:
                    vote_d[i] = {}

                k = get_option_list_choice(new_d[i])[0]
                k1 = get_option_list_choice(old_d[i])[0]
                sub_ballot_d = The.preset.get_vote_d(k)

                # A Frame OptionList has no vote in
                # the ballot dict, so it is checked here.
                if new_d[i][ok.SWITCH] != old_d[i][ok.SWITCH]:
                    reap_ballot(vote_d[i], sub_ballot_d, new_d[i][k])

                elif k == k1:
                    reap_vote(
                        vote_d[i], sub_ballot_d, old_d[i][k], new_d[i][k]
                    )
                else:
                    reap_ballot(vote_d[i], sub_ballot_d, new_d[i][k])

            elif isinstance(a, dict):
                if i not in vote_d:
                    vote_d[i] = {}
                reap_vote(vote_d[i], a, old_d[i], new_d[i])
            elif old_d[i] != new_d[i]:
                accept_vote(vote_d, i, a, True)
    else:
        reap_ballot(vote_d, ballot_d, new_d)


def reap_ballot(vote_d, ballot_d, new_d):
    """
    Collect vote from the default ballot dict.

    vote_d: dict
        Use to collect vote.

    ballot_d: dict
        Is the default vote for a Preset.
    """
    for i, a in ballot_d.items():
        if i == ok.FRAME:
            k = get_option_list_choice(new_d[i])[0]
            sub_ballot_d = The.preset.get_vote_d(k)

            if i not in vote_d:
                vote_d[i] = {}
            reap_ballot(vote_d[i], sub_ballot_d, new_d[i][k])

        elif isinstance(a, dict):
            if i not in vote_d:
                vote_d[i] = {}
            reap_ballot(vote_d[i], a, new_d[i])
        else:
            accept_vote(vote_d, i, a, True)


def update_button_visual(g):
    """
    Update the Open Button's visibility.

    g: Per Cell Group
        Has button.
    """
    if g.check_button.get_a():
        g.button.show()
    else:
        g.button.hide()


class PerCellGroup(gtk.Alignment, gobject.GObject, object):
    """
    Has a CheckButton and a Button. The group
    value is a 2D list containing a cell table.
    """
    __gsignals__ = si.PER_CELL_D
    change_signal = None
    has_table_label = True

    def __init__(self, **d):
        """
        Create Widgets for switching to the Per Cell option.

        d: dict
            Has keyword argument for Widget.
        """
        super(gtk.Alignment, self).__init__()
        gobject.GObject.__init__(self)

        # the GTK Alignment setting
        self.set(0, 0, 1, 1)

        self.handle_d = {}
        self.default_ballot_d = None
        self._cancel_value = {}
        self._hbox = gtk.HBox()

        # Updated after a View, a copy of the value table.
        # Is ordered as [Plan, Work].
        self.view_value = [{}, {}]

        # Is a 2D table when it has cell data. The cell data
        # can be a tuple for a Table's Cell Type, or a Preset
        # for any other group type.
        self._value = {}

        # Is true when PortCellTable is open.
        self.edit_mode = False

        set_widget_attr(self, d)

        # Use to clear the Per Cell value table on a Preset load.
        self.is_empty = self.any_group.render_key in sk.NO_PER_CELL

        if not self.is_empty:
            self.default_ballot_d = The.preset.get_vote_d(
                self.any_group.item.key
            )

        self.dialog = PortPerCell
        relay = d[wk.RELAY][:]
        same_size = gtk.SizeGroup(mode=gtk.SIZE_GROUP_BOTH)

        # CheckButton
        # The CheckButton and Button don't cast vote.
        d.pop(wk.ISSUE)
        d[wk.RELAY].insert(0, self.on_per_cell_switch)

        d[wk.CHANGELESS] = True
        d[wk.TOOLTIP] = Tip.PER_CELL_CHECK_BUTTON
        self.is_type_table = self.any_group.item.key == gk.TYPE_TABLE
        d[wk.TEXT] = "Merge Cell" if self.is_type_table else "Per Cell"
        d[wk.ALIGN] = 0, 0, 0, 1
        self.check_button = CheckButton(**d)

        # Button
        d[wk.TOOLTIP] = Tip.PER_CELL_BUTTON
        d[wk.KEY] = d[wk.TEXT] = bk.OPEN
        d[wk.ALIGN] = 0, 0, 1, 0
        d[wk.RELAY] = relay

        relay.insert(0, self.on_open_button_action)

        self.button = self.widget = Button(**d)

        self.add(self._hbox)

        for i in (self.check_button, self.button):
            self._hbox.add(i)
            same_size.add_widget(i.widget)

        baby = self.any_group.item.model.baby
        if not self.is_type_table:
            self.handle_d[baby.connect(
                si.CELL_RECT_CALC, self.on_cell_rect_calc
            )] = baby

    def get_a(self):
        """
        Get the Preset table.

        Return: list
            Preset table
            2D (list(s) within a list)
            May be an empty list.
        """
        return self._value

    def get_cell_value(self, r_c):
        """
        Fetch the value from the Preset table.

        r_c: tuple
            (r, c); cell index
        """
        if r_c in self._value:
            return self._value[r_c]

    def hide(self, *_):
        """
        Hide the row in the Table container if the Widget is ready.
        Table sets the 'self.box' to a VBox container.
        """
        if self.box:
            self.box.hide()
            self.label_box.hide()

    def intersect_vote(self, vote_d):
        """
        Receive a vote dict and send a change signal for each cell.

        vote_d: dict
            {(r, c): [Plan vote, Work vote]}
        """
        group_vote_q = set()
        q = self._value.keys()

        # Plan and Work vote list, 'a'
        for r_c, a in vote_d.items():
            if r_c in q:
                # Plan and Work, '2'
                for x in range(2):
                    d = {ok.IS_CHAIN: a[x]}
                    self.emit(
                        si.PER_CELL_CHANGE,
                        (r_c, {vo.MATTER: d, vo.CAKE: d}, x)
                    )
                    group_vote_q.update((x,))
        for x in group_vote_q:
            self.any_group.cast_vote(x, ok.PER_CELL, vo.PER_CELL, True)

    def on_accept_cell_edit(self, d):
        """
        Return from PortCell. Exit from edit mode.

        d: dict
            the new value for the PerCellGroup
        """
        self.set_a(d)
        self.edit_mode = False

        # When the Table Cell Type change,
        # the pocket and merge rect need update.
        if self.is_type_table:
            self.any_group.item.model.baby.give(
                si.CELL_RECT_CHANGE, (self.any_group.value_d, False)
            )

    def on_cancel_cell_edit(self):
        """
        Return from PortCell. Exit from edit
        mode. Restore the 'cancel_value' table.
        """
        self.set_a(self._cancel_value)
        self.edit_mode = False
        if self.is_type_table:
            # Merge cell may have been altered.
            self.any_group.item.model.baby.give(
                si.CELL_RECT_CHANGE, (self.any_group.value_d, False)
            )

    def on_cell_rect_calc(self, _, arg):
        """
        Check the cell key for a Model grid.

        _: Baby
            Sent the Signal.

        arg: dict
            a vote dict
        """
        lost_q = []
        is_change = False
        q = self.any_group.item.model.cell_q

        for k in self._value.keys():
            if k not in q:
                lost_q.append(k)
                self._value.pop(k)
            else:
                # The Per Maya changed.
                is_change = True

        if is_change:
            # Set the View value to trigger change.
            self.view_value = [{}, {}]
            compare_both_value(self)
        if lost_q:
            self.emit(si.DISAPPEAR, lost_q)


    def on_open_button_action(self, g):
        """
        Respond to an Open Button action.

        g: Button
            Is responsible.
        """
        # Use to restore change on cancel.
        self._cancel_value = deepcopy(self._value)

        # Is True when the cell editor is open.
        self.edit_mode = True

        self.roller_win.bring_dialog(
            self,
            d={
                wk.ON_ACCEPT: self.on_accept_cell_edit,
                wk.ON_CANCEL: self.on_cancel_cell_edit,
            }
        )

    def on_per_cell_switch(self, g):
        """
        Respond to a change event for the Per Cell CheckButton.

        g: CheckButton
            Is responsible.
        """
        if not self.is_empty:
            if g.get_a():
                if self.is_type_table and not The.load_count:
                    for r_c in self.any_group.item.model.cell_q:
                        self._value[r_c] = 1, 1
            else:
                self.set_a({})

        # Plan and Work indices, '2'
        for x in range(2):
            self.any_group.cast_vote(
                x, ok.PER_CELL, vo.PER_CELL, self._value != self.view_value[x]
            )
        update_button_visual(self)

    def set_a(self, d, x=None):
        """
        Set the PerCellGroup value.

        d: dict
            {(row, column): Preset value dict}
            The row and column cell index range from 0
            to the span of the Model's table.

        v: View or None
            Has scope variable. Use when fixing tables while in
            a View process. Determine which preview value table
            is compared depending on the View origin.

        x: int or None
            Is used by Cell Table to indicate the View origin
            where zero is Plan and one is Work.
        """
        if not self.is_empty:
            last_value = deepcopy(self._value)
            d = self._value = d if isinstance(d, dict) else {}

            if x is not None:
                compare_value_table(self, x)

            else:
                compare_both_value(self)
            if not self.is_type_table:
                lost_q = []

                for r_c in last_value:
                    if r_c not in self._value:
                        lost_q.append(r_c)
                if lost_q:
                    self.emit(si.DISAPPEAR, lost_q)

        self.check_button.widget.set_active(bool(d))
        self.any_group.update_option_a(self.key, d)

    def set_view_value(self, x, a):
        self.view_value[x] = deepcopy(a)

    def show(self, *_):
        """
        Show the row in the Table Widget if the Widget is ready.
        The Table Widget class has set the 'self.box' to a VBox container.
        """
        if self.box:
            self.box.show()
            self.label_box.show()


# Register the custom signals.
gobject.type_register(PerCellGroup)
